"""A parser to handle propositional phrases.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

A proposition is an un unevaluated logical expression that can be true
or false, depending on the environment we evaluate one in.

At the moment, I'm reading "The Science Of Programming", by David
Gries.  He has a few chapters on propositional logic, and I got
inspired enough to write a Python program to work with propositional
phrases.

Here are examples of some propositions:

    not X
    x and y and z
    x and y or (x and y)

The last should be interpreted as:

    (x and y) or (x and y)

That is, 'and' should bind more tightly than 'or'.  'not' binds the most
tightly.  That's the sort of knowledge I'll try to build into the parser.

Here's the list of Expression, in order of increasing precedence.

    Expression
    EqualityExpression
    ImplicationExpression
    OrExpression
    AndExpression
    NotExpression
    AtomicExpression
"""

import spark
import structure

class ParserError(Exception): pass


def parse(s):
    """Given a string, returns an AST of the parsed expression."""
    import lexer
    p = ExpressionParser()
    try:
        tokens = lexer.tokenize(s)
        return p.parse(tokens)
    except spark.ParserError, e:
        raise ParserError, e


class ExpressionParser(spark.GenericParser):
    def __init__(self, start='Expression'):
        spark.GenericParser.__init__(self, start)

    def p_Expression(self, args):
        ' Expression ::= EqualityExpression '
        return args[0]

    def p_EqualityExpression1(self, args):
        ''' EqualityExpression ::= EqualityExpression
                                       equals ImplicationExpression '''
        return structure.EqualityNode(args[0], args[2])

    def p_EqualityExpression2(self, args):
        ' EqualityExpression ::= ImplicationExpression '
        return args[0]

    def p_ImplicationExpression1(self, args):
        ''' ImplicationExpression ::= ImplicationExpression
                                          implies OrExpression '''
        return structure.ImplicationNode(args[0], args[2])

    def p_ImplicationExpression2(self, args):
        ' ImplicationExpression ::= OrExpression '
        return args[0]

    def p_OrExpression1(self, args):
        ' OrExpression ::= OrExpression or AndExpression '
        return structure.OrNode(args[0], args[2])

    def p_OrExpression2(self, args):
        ' OrExpression ::= AndExpression '
        return args[0]

    def p_AndExpression1(self, args):
        ' AndExpression ::= AndExpression and NotExpression '
        return structure.AndNode(args[0], args[2])

    def p_AndExpression2(self, args):
        ' AndExpression ::= NotExpression '
        return args[0]

    def p_NotExpression1(self, args):
        ' NotExpression ::= not NotExpression '
        return structure.NotNode(args[1])

    def p_NotExpression2(self, args):
        ' NotExpression ::= AtomicExpression '
        return args[0]

    def p_atomicExpression1(self, args):
        ' AtomicExpression ::= identifier '
        return structure.IdentifierNode(args[0].attr)

    def p_atomicExpression2(self, args):
        ' AtomicExpression ::= constant '
        value = args[0].attr
        if value == structure.TRUE_VALUE:
            return structure.TrueNode()
        return structure.FalseNode()

    def p_atomicExpression3(self, args):
        ' AtomicExpression ::= ( Expression ) '
        return args[1]


def test():
    import prompt
    print """This is a small test of the propositional parser.
Enter propositional phrases at the prompt.  To quit, type "quit" at
the prompt.

Examples of some phrases:

    jack and jill
    rock and rolling implies not moss
"""
    p = prompt.Prompt('ParserTest', callback=_promptCallback)
    p.promptLoop()

def _promptCallback(line):
    try:
        tree = parse(line)
        if tree: print tree
        print
    except ParserError, e:
        print '''There was a problem parsing your line:\n%s''' % e
        print
        
if __name__ == '__main__':
    test()
